/***************************************************************************
 *   Copyright (C) 2007  Antony Martin <martin.antony@yahoo.fr>            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU General Public License           *
 *   as published by the Free Software Foundation; either version 2        *
 *   of the License, or (at your option) any later version.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************/

/* Cree le : 2 juillet 2007
   derniere modification le 03/07/2007 */


#include <SCE/SCEMinimal.h>

#include <SCE/core/SCECTexture.h>

#include <SCE/core/SCECFramebuffer.h>



/* ajoute le 02/07/2007 */
SCE_CFramebuffer* SCE_CCreateFramebuffer(void)
{
    SCE_CFramebuffer *fb = NULL;
    unsigned int i;
    
    
    fb = SCE_malloc(sizeof *fb);
    if(fb == NULL)
    {
        Logger_LogSrc();
        return (NULL);
    }
    
    for(i=0; i<SCE_NUM_RENDER_BUFFERS; i++)
    {
        fb->buffers[i].tex = NULL;
        fb->buffers[i].id = 0;
        fb->buffers[i].user = SCE_TRUE;
        fb->buffers[i].clear = SCE_FALSE;
    }
    
    fb->x = fb->y = fb->w = fb->h = 0;
    
    glGenFramebuffersEXT(1, &fb->id);
    
    fb->mask = 0;
    
    return (fb);
}

/* ajoute le 02/07/2007 */
void SCE_CDeleteFramebuffer(SCE_CFramebuffer *fb)
{
    unsigned int i;
    
    if(fb != NULL)
    {
        for(i=0; i<SCE_NUM_RENDER_BUFFERS; i++)
        {
            if(fb->buffers[i].user)
                SCE_CDeleteTexture(fb->buffers[i].tex);
            
            if(fb->buffers[i].id)
                glDeleteRenderbuffersEXT(1, &fb->buffers[i].id);
        }
        
        glDeleteFramebuffersEXT(1, &fb->id);
        
        SCE_free(fb);
    }
}



/* converti un simple ID de type d'attachement en type opengl */
static int SCE_CIDToGLBuffer(SCEuint type)
{
    if(type == SCE_COLOR_BUFFER)
        return GL_COLOR_ATTACHMENT0_EXT;
    else if(type == SCE_DEPTH_BUFFER)
        return GL_DEPTH_ATTACHMENT_EXT;
    else
        return GL_STENCIL_ATTACHMENT_EXT;
}

/* ajoute le 02/07/2007 */
/* revise le 03/07/2007 */
void SCE_CAddRenderTexture(SCE_CFramebuffer *fb, SCEuint id,
                           SCEenum target, SCE_CTexture *tex, int mipmap)
{
    int type = SCE_CIDToGLBuffer(id), i;
    
    
    if(fb != NULL)
    {
        /* si la precedente texture n'a pas ete envoyee par l'utilisateur,
           on la supprime */
        if(!fb->buffers[id].user)
        {
            SCE_CDeleteTexture(fb->buffers[id].tex);
            fb->buffers[id].tex = NULL;  /* ! important ! */
        }
        
        /* si un buffer existe deja, on le supprime */
        if(fb->buffers[id].id)
        {
            glDeleteRenderbuffersEXT(1, &fb->buffers[id].id);
            fb->buffers[id].id = 0;
        }
        
        fb->buffers[id].tex = tex;
        
        if(target == 0)
            target = tex->target;
        
        /* on bind le FBO */
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fb->id);
        /* et on lui assigne notre zolie render texture */
        glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, type,
                                  target, tex->id, mipmap);
        
        /* si aucune color render texture n'existe, on desactive le tampon
           (me demandez pas pourquoi, j'en sais rien, allez faire un tour ici :
            http://www.cs.cmu.edu/afs/cs/academic/class/15462/web.06s/
            asst/project3/shadowmap/fbo.html
        */
        if(id == SCE_DEPTH_BUFFER && fb->buffers[SCE_COLOR_BUFFER].tex == NULL)
            glDrawBuffer(GL_NONE);
        
        /* si en revanche un cree une color render texture, et qu'une depth
           render texture est deja creee, on reactive le tampon (logique la) *//*
        if(id == SCE_COLOR_BUFFER)
        {
            glDrawBuffer(GL_BACK);
        }*/ /* TODO: gros fuck.... */
        
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
        
        /* si le mipmap n'a pas ete specifie */
        if(!tex->spec_mipmap)
        {
            /* NOTE: et dans le cas d'une cubemap, on ne choppe les
                     dimensions que de POSX... */
            /* on ne peut acceder qu'au premier element */
            fb->w = tex->data[0].w * 2; /* hack de merde pour un for de merde */
            fb->h = tex->data[0].h * 2;
            
            /* calcul des nouvelles dimensions */
            for(i=mipmap; i>=0; i--) /* <-- le for en question... */
            {
                fb->w /= 2;
                fb->h /= 2;
            }
        }
        else if(!tex->texcub)
        {
            fb->w = tex->data[mipmap].w;
            fb->h = tex->data[mipmap].h;
        }
        else
        {
            /* dans le cas d'une cubemap, on choppe les dimensions
               de la bonne texture correspondante au target */
            /* pour chaque donnee... */
            for(i=0; i<(int)tex->n_datas; i++)
            {
                /* ...on l'analyse */
                if(tex->data[i].target == target)
                {
                    fb->w = tex->data[i].w;
                    fb->h = tex->data[i].h;
                    break;
                }
            }
        }
        
        SCE_CClearRenderbuffer(fb, id, 1);
    }
}

/* ajoute le 02/07/2007 */
/* revise le 03/07/2007 */
void SCE_CAddRenderBuffer(SCE_CFramebuffer *fb, SCEuint id,
                          int fmt, int w, int h)
{
    int type = SCE_CIDToGLBuffer(id);
    
    if(fb != NULL)
    {
        /* assignation des valeurs par defaut */
        if(fmt == 0)
        {
            if(id == SCE_DEPTH_BUFFER)
                fmt = GL_DEPTH_COMPONENT24;
            else if(id == SCE_STENCIL_BUFFER)
                fmt = GL_STENCIL_INDEX8_EXT;
        }
        
        /* si une texture existait deja, on la supprime */
        if(!fb->buffers[id].user)
        {
            SCE_CDeleteTexture(fb->buffers[id].tex);
            fb->buffers[id].tex = NULL;  /* ! important ! */
        }
        
        /* si un buffer existe deja, on le supprime */
        if(fb->buffers[id].id)
        {
           glDeleteRenderbuffersEXT(1, &fb->buffers[id].id);
           fb->buffers[id].id = 0;
        }
        
        /* creation du render buffer */
        glGenRenderbuffersEXT(1, &fb->buffers[id].id);
        glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, fb->buffers[id].id);
        glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, fmt, w, h);
        glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, 0);
        
        /* on l'ajoute au FBO */
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fb->id);
        glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, type,
                                     GL_RENDERBUFFER_EXT, fb->buffers[id].id);
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
        
        fb->w = w;
        fb->h = h;
        
        SCE_CClearRenderbuffer(fb, id, 1);
    }
}

/* ajoute le 02/07/2007 */
/* revise le 03/07/2007 */
int SCE_CCreateRenderTexture(SCE_CFramebuffer *fb, SCEuint id,
                             int internal, int format, int type, int w, int h)
{
    SCE_CTexture *tex = NULL;
    SCE_CTexData data;
    
    
    /* creation de la texture */
    tex = SCE_CCreateTexture(SCE_TEX_2D);
    if(tex == NULL)
    {
        Logger_LogSrc();
        return (SCE_ERROR);
    }
    
    /* utilisation de valeurs par defaut en cas de non specification */
    if(internal == 0)
    {
        if(id == SCE_DEPTH_BUFFER)
            internal = GL_DEPTH_COMPONENT24;
        else
            internal = GL_RGBA;
    }
    if(format == 0)
    {
        if(id == SCE_DEPTH_BUFFER)
            format = GL_DEPTH_COMPONENT;
        else
            format = GL_RGBA;
    }
    if(type == 0)
        type = SCE_UNSIGNED_BYTE;
    
    data.level = 0;
    data.w = w;
    data.h = h;
    data.type = type;
    data.internal = internal;
    data.format = format;
    data.texels = NULL;
    
    SCE_CAssociateTextureData(tex, &data);
    
    if(SCE_CAddTextureData(tex, &data) < 0)
    {
        Logger_LogSrc();
        SCE_CDeleteTexture(tex);
        return (SCE_ERROR);
    }
    if(SCE_CBuildTexture(tex, 0, 0) < 0)
    {
        Logger_LogSrc();
        SCE_CDeleteTexture(tex);
        return (SCE_ERROR);
    }
    /***/
    
    /* ajout de la texture en tant que render buffer */
    SCE_CAddRenderTexture(fb, id, SCE_TEX_2D, tex, 0);
    
    /* attention: la texture n'appartient pas a l'utilisateur */
    fb->buffers[id].user = SCE_FALSE;
    
    return (SCE_OK);
}

/* ajoute le 03/07/2007 */
/* NOTE: mouais... assez moyen comme fonction */
SCE_CTexture* SCE_CGetRenderTexture(SCE_CFramebuffer *fb, SCEuint id)
{
    return (fb->buffers[id].tex);
}


/* ajoute le 02/07/2007 */
void SCE_CClearRenderbuffer(SCE_CFramebuffer *fb, SCEuint type, int active)
{
    if(fb != NULL)
    {
        switch(type)
        {
            case SCE_COLOR_BUFFER:
                if(active)
                    fb->mask |= GL_COLOR_BUFFER_BIT;
                else
                    fb->mask -= GL_COLOR_BUFFER_BIT;
                break;
            
            case SCE_DEPTH_BUFFER:
                if(active)
                    fb->mask |= GL_DEPTH_BUFFER_BIT;
                else
                    fb->mask -= GL_DEPTH_BUFFER_BIT;
                break;
            
            case SCE_STENCIL_BUFFER:
                if(active)
                    fb->mask |= GL_STENCIL_BUFFER_BIT;
                else
                    fb->mask -= GL_STENCIL_BUFFER_BIT;
                break;
        }
    }
}


/* ajoute le 02/07/2007 */
/* ajoute le 07/07/2007 */
void SCE_CUseFramebuffer(SCE_CFramebuffer *fb, SCE_CRect *r)
{
    static int viewport[4];     /* viewport avant l'utilisation du fbo */
    static int binded = 0;      /* booleen, true: un fbo est deja binde */
    
    if(fb != NULL)
    {
        /* on recupere d'abord le viewport, pour ensuite le restituer
           (uniquement si aucun fbo n'est deja binde,
            sinon on prendrait son viewport) */
        if(!binded)
            glGetIntegerv(GL_VIEWPORT, viewport);
        
        if(r != NULL)
            glViewport(r->x, r->y, r->w, r->h);
        else
            glViewport(fb->x, fb->y, fb->w, fb->h);
        
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fb->id);
        glClear(fb->mask);
        
        binded = SCE_TRUE;
    }
    else if(binded)
    {
        glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
        glViewport(viewport[0], viewport[1], viewport[2], viewport[3]);
        binded = SCE_FALSE;
    }
}
