/***************************************************************************
 *   Copyright (C) 2007  Antony Martin <martin.antony@yahoo.fr>            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU General Public License           *
 *   as published by the Free Software Foundation; either version 2        *
 *   of the License, or (at your option) any later version.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************/

/* Cree le : 14 janvier 2007
   derniere modification le 10/07/2007 */



#include <SCE/SCEMinimal.h>

#include <SCE/utils/SCEString.h>
#include <SCE/utils/SCEMath.h>

#include <SCE/core/SCECSupport.h>

#include <SCE/core/SCECTexture.h>


/* constante(s) interne(s) */
/* dimensions du mipmapping le plus eleve */
#define SCE_MIN_MIPMAP_DIM 2


/* ajotue le 10/07/2007 */
/* texture actuellement bindee */
static SCE_CTexture *binded = NULL;


SCE_CTexture* SCE_CCreateTexture(SCEenum target)
{
    SCE_CTexture *tex = NULL;
    
    tex = SCE_malloc(sizeof *tex);
    if(tex == NULL)
    {
        Logger_LogSrc();
        return (NULL);
    }
    
    tex->tex1d = tex->tex2d = tex->tex3d = tex->texcub = 0;
    
    if(SCE_CSetTextureTarget(tex, target) < 0)
    {
        SCE_free(tex);
        Logger_LogSrc();
        return (NULL);
    }
    
    tex->base_level = 0;
    tex->max_level = 0;
    
    tex->spec_mipmap = 0;
    tex->use_mipmap = 0;
    tex->hw_mipmap = 0;
    tex->have_data = 0;
    
    tex->builded = 0;
    
    tex->n_datas = 0;
    tex->data = NULL;
    
    glGenTextures(1, &tex->id);
    
    /* quelques parametres */ /* NOTE: NON !
    (on laisse l'utilisateur se demerder avec ses parametres :-) *//*
    glBindTexture(GL_TEXTURE_2D, tex->id);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    glBindTexture(GL_TEXTURE_2D, 0);*/
    
    return (tex);
}

void SCE_CDeleteTexture(SCE_CTexture *tex)
{
    if(tex)
    {
        glDeleteTextures(1, &tex->id);
        
        if(SCE_free(tex->data) < 0)
        {
            /* NOTE: c'est pitetre con de logguer une erreur pour ca... ? */
            Logger_LogSrc();
        }
        
        if(SCE_free(tex) < 0)
        {
            /* NOTE: c'est pitetre con de logguer une erreur pour ca... ? */
            Logger_LogSrc();
        }
    }
}

int SCE_CSetTextureTarget(SCE_CTexture *tex, SCEenum target)
{
    switch(target)
    {
        case SCE_TEX_1D:
            tex->tex1d = 1;
            break;
        
        case SCE_TEX_2D:
            tex->tex2d = 1;
            break;
        
        case SCE_TEX_3D:
            tex->tex3d = 1;
            break;
        
        case SCE_TEX_CUBE:
            tex->texcub = 1;
            tex->tex2d = 1;
            break;
        
        #ifdef __SCE_DEBUG__
        default:
            Logger_Log(SCE_INVALID_ARG);
            Logger_LogMsg("target is not a valid texture type");
            return (SCE_ERROR);
        #endif
    }
    
    tex->target = target;
    
    return (SCE_OK);
}

/*
static unsigned int SCE_CGetPixelSize(GLenum format)
{
    unsigned int size;
    switch(format)
    {
        case GL_ALPHA4:
        case GL_LUMINANCE4:
        case GL_INTENSITY4:
        case GL_:
            size = 4;
            break;
        
        case GL_COMPRESSED_RGBA_S3TC_DXT1:
        case GL_ALPHA8:
        case GL_ALPHA:
        case GL_LUMINANCE:
        case GL_LUMINANCE8:
        case GL_LUMINANCE4_ALPHA4:
        case GL_LUMINANCE6_ALPHA2:
        case GL_INTENSITY:
        case GL_INTENSITY8:
        case GL_R3_G3_B2:
        case GL_RGBA2:
            size = 8;
            break;
        
        case GL_ALPHA12:
        case GL_LUMINANCE12:
        case GL_INTENSITY12:
        case GL_RGB4:
            size = 12;
            break;
        
        case GL_COMPRESSED_RGBA_S3TC_DXT3:
        case GL_COMPRESSED_RGBA_S3TC_DXT5:
        case GL_LUMINANCE16:
        case GL_LUMINANCE_ALPHA:
        case GL_LUMINANCE8_ALPHA8:
        case GL_INTENSITY16:
        case GL_INTENSITY12_ALPHA4:
        case GL_RGBA4:
        case GL_ALPHA16:
        case GL_RGB5_A1:
            size = 16;
            break;
        
        case GL_RGB8:
        case GL_RGB:
        case GL_LUMINANCE12_ALPHA12:
            size = 24;
            break;
        
        case GL_RGB10:
            size = 30;
            break;
        
        case GL_RGBA:
        case GL_RGBA8:
        case GL_LUMINANCE16_ALPHA16:
        case GL_RGB10_A2:
            size = 32;
            break;
        
        case GL_RGB12:
            size = 36;
            break;
        
        case GL_RGB16:
        case GL_RGBA12:
            size = 48;
            break;
        
        case GL_RGBA16:
            size = 64;
            break;
        
        default:
            size = 0;
    }
    
    return (size);
}*/


int SCE_CCopyTextureData(SCE_CTexture *tex, unsigned int id, SCE_CTexData *data)
{
    #ifdef __SCE_DEBUG__
    if(tex == NULL)
    {
        Logger_Log(SCE_INVALID_ARG);
        return (SCE_ERROR);
    }
    #endif
    
    memcpy(&tex->data[id], data, sizeof *data);
    
    return (SCE_OK);
}

int SCE_CAssociateTextureData(SCE_CTexture *tex, SCE_CTexData *data)
{
    #ifdef __SCE_DEBUG__
    if(tex == NULL || data == NULL)
    {
        Logger_Log(SCE_INVALID_ARG);
        return (SCE_ERROR);
    }
    #endif
    
    if(data->target != SCE_TEX_POSX &&
       data->target != SCE_TEX_NEGX &&
       data->target != SCE_TEX_POSY &&
       data->target != SCE_TEX_NEGY &&
       data->target != SCE_TEX_POSZ &&
       data->target != SCE_TEX_NEGZ)
    {
        /* si le 'target' ne specifie aucune position cube map,
           on lui affecte la valeur de target ar defaut de la texture */
        data->target = tex->target;
    }
    
    if(data->internal == GL_COMPRESSED_RGBA_S3TC_DXT1_EXT ||
       data->internal == GL_COMPRESSED_RGBA_S3TC_DXT3_EXT ||
       data->internal == GL_COMPRESSED_RGBA_S3TC_DXT5_EXT ||
       data->internal == GL_COMPRESSED_RGB_3DC_ATI)
    {
        data->comp = 1;
    }
    else
        data->comp = 0;
    
    /* la taille n'est utile que pour les formats compresses */
    if(data->comp)
    {
        /* taille d'un pixel (en bytes) */
        data->pxsize =
        (data->internal == GL_COMPRESSED_RGBA_S3TC_DXT1_EXT) ? 8 : 16;
        /* NOTE: et 3Dc ? */
        
        /* calculs foireux mais qui sont soit-disant justes... */
        if(tex->tex1d)
            data->size =
            ((data->w+3)/4) * data->pxsize;
        else if(tex->tex2d || tex->texcub)
            data->size =
            ((data->w+3)/4) * ((data->h+3)/4) * data->pxsize;
        else
            data->size =
            ((data->w+3)/4) * ((data->h+3)/4) * ((data->d+3)/4) * data->pxsize;
    }
    
    return (SCE_OK);
}

int SCE_CAddTextureData(SCE_CTexture *tex, SCE_CTexData *data)
{
    #ifdef __SCE_DEBUG__
    if(tex == NULL)
    {
        Logger_Log(SCE_INVALID_ARG);
        return (SCE_ERROR);
    }
    #endif
    
    
    /* creation d'une nouvelle donnee */
    tex->data = SCE_realloc(tex->data, (tex->n_datas+1) * sizeof *tex->data);
    if(tex->data == NULL)
    {
        Logger_LogSrc();
        return (SCE_ERROR);
    }
    
    /* copie des donnees */
    SCE_CCopyTextureData(tex, tex->n_datas, data);
    
    /* des donnees ont ete envoyees, booleen positionne a 'true' */
    tex->have_data = 1;
    
    /* et 'une' donnee de plus */
    tex->n_datas++;
    
    return (SCE_OK);
}

int SCE_CLocateTextureDataID(SCE_CTexture *tex, SCEenum target, int level)
{
    unsigned int i;
    
    #ifdef __SCE_DEBUG__
    if(tex == NULL)
    {
        Logger_Log(SCE_INVALID_ARG);
        return (SCE_ERROR);
    }
    #endif
    
    if(target == SCE_TEX_POSX ||
       target == SCE_TEX_NEGX ||
       target == SCE_TEX_POSY ||
       target == SCE_TEX_NEGY ||
       target == SCE_TEX_POSZ ||
       target == SCE_TEX_NEGZ)
    {
        for(i=0; i<tex->n_datas; i++)
        {
            if(tex->data[i].target == target)
                return (i);
        }
    }
    else
    {
        for(i=0; i<tex->n_datas; i++)
        {
            if(tex->data[i].level == level)
                return (i);
        }
    }
    
    return (-1);
}

int SCE_CUpdateTextureData(SCE_CTexture *tex, int id, SCE_CTexData *data)
{
    #ifdef __SCE_DEBUG__
    if(tex == NULL || data == NULL)
    {
        Logger_Log(SCE_INVALID_ARG);
        return (SCE_ERROR);
    }
    if(id < 0 || (unsigned int)id > tex->n_datas)
    {
        Logger_Log(SCE_INVALID_ARG);
        return (SCE_ERROR);
    }
    #endif
    
    SCE_CCopyTextureData(tex, id, data);
    
    return (SCE_OK);
}

/* NOTE: nom de fonction a revoir... encore que. */
void SCE_CDeleteTextureData(SCE_CTexture *tex)
{
    if(tex != NULL)
    {
        /* on s'assure que la suppression a egalement lieue cote serveur */
        glDeleteTextures(1, &tex->id);
        
        /* puis on re-cree l'identifiant */
        glGenTextures(1, &tex->id);
        
        /* enfin, on supprime les donnees */
        SCE_free(tex->data);
        tex->data = NULL;
        
        /* ne pas oublier de replacer ces gentilles petites valeurs a zero */
        tex->n_datas = tex->builded = tex->have_data = tex->spec_mipmap =
        tex->hw_mipmap = tex->use_mipmap = 0;
    }
}


/* revise le 10/07/2007 */
void SCE_CSetTextureFilter(SCE_CTexture *tex, SCEint level)
{
    if(tex != NULL)
    {
        tex->min_filter = level;
        
        /* MAG_FILTER ne peut etre que nearest ou linear */
        if(tex->min_filter == GL_NEAREST)
            tex->mag_filter = GL_NEAREST;
        else
            tex->mag_filter = GL_LINEAR;
        
        glBindTexture(tex->target, tex->id);
        glTexParameteri(tex->target, GL_TEXTURE_MIN_FILTER, tex->min_filter);
        glTexParameteri(tex->target, GL_TEXTURE_MAG_FILTER, tex->mag_filter);
        glBindTexture(tex->target, 0);
    }
}

/* revise le 10/07/2007 */
void SCE_CTextureParam(SCEenum pname, SCEint param)
{
    if(binded != NULL)
        glTexParameteri(binded->target, pname, param);
}

/* revise le 10/07/2007 */
void SCE_CTextureParamf(SCEenum pname, SCEfloat param)
{
    if(binded != NULL)
        glTexParameterf(binded->target, pname, param);
}

/* ajoute le 10/07/2007 */
void SCE_CTextureGen(SCEenum a, SCEenum b, int c)
{
    glTexGeni(a, b, c);
}

/* ajoute le 10/07/2007 */
void SCE_CTextureGenf(SCEenum a, SCEenum b, float c)
{
    glTexGenf(a, b, c);
}

/* ajoute le 10/07/2007 */
void SCE_CTextureGenfv(SCEenum a, SCEenum b, float *c)
{
    glTexGenfv(a, b, c);
}


/* revise le 04/07/2007 */
int SCE_CBuildTexture(
  SCE_CTexture *tex,    /* texture */
  int use_mipmap,       /* booleen, true: generation automatique de mipmaps */
  int hw_mipmap)        /* booleen, true: generation hardware du mipmap */
{
    /* compteurs */
    unsigned int i=0, j;
    
    /* tours de boucle pour envoyer les images cube map (si presentes) */
    unsigned int npass=1;
    
    /* dimensions temporaires de l'image (mipmapping software) */
    int w, h, d;
    
    /* taille temporaire */
    unsigned int size;
    
    #ifdef __SCE_DEBUG__
    if(tex == NULL)
    {
        Logger_Log(SCE_INVALID_ARG);
        return (SCE_ERROR);
    }
    #endif
    
    /* si la texture a deja ete construite */
    if(tex->builded)
    {
        #ifdef __SCE_DEBUG__
        Logger_PrintMsg("you can't rebuild a texture, use SCE_CUpdateTexture");
        #endif
        return (SCE_OK);
    }
    
    /* si la texture ne contient aucune donnee */
    if(tex->have_data == 0)
    {
        Logger_Log(SCE_INVALID_OPERATION);
        Logger_LogMsg("you must specify data before build texture");
        return (SCE_ERROR);
    }
    
    /* si plusieurs donnees ont ete specifiees, l'utilisateur
       reclame 'obligatoirement' l'utilisation du mipmapping */
    if(tex->target != SCE_TEX_CUBE)
    {
        if(tex->n_datas > 1)
            tex->spec_mipmap = 1;
    }
    else if(tex->n_datas > 6)
        tex->spec_mipmap = 1;
    
    if(use_mipmap || tex->spec_mipmap)
        tex->use_mipmap = 1;
    
    /* verification du support de la generation en hardware du mipmapping
       on verifie egalement si l'utilisateur a envoye manuellement
       tous les niveaux de mipmapping (si c'est le cas, on ne genere
       pas de mipmap en hardware... :/ ca n'aurait aucun sens ) */
    if(!SCE_CHasCap(SCE_TEX_HW_GEN_MIPMAP) || tex->spec_mipmap)
        hw_mipmap = 0;
    
    tex->hw_mipmap = hw_mipmap;
    
    
    /* macro barbare, mais bien pratique :) */
    /* NOTE: est egalement utilise dans CUpdateTexture */
    #define SCE_TEXIMAGE(target, level, internal, w, h, d,\
                         size, format, type, texels, comp)\
    {\
        if(comp)\
        {\
            if(tex->tex1d)\
                glCompressedTexImage1D(target, level, internal,\
                                       w, 0, size, texels);\
            else if(tex->tex2d)\
                glCompressedTexImage2D(target, level, internal,\
                                       w, h, 0, size, texels);\
            else if(tex->tex3d)\
                glCompressedTexImage3D(target, level, internal,\
                                       w, h, d, 0, size, texels);\
        }\
        else\
        {\
            if(tex->tex1d)\
                glTexImage1D(target, level, internal,\
                             w, 0, format, type, NULL);\
            else if(tex->tex2d)\
                glTexImage2D(target, level, internal,\
                             w, h, 0, format, type, NULL);\
            else if(tex->tex3d)\
                glTexImage3D(target, level, internal,\
                             w, h, d, 0, format, type, NULL);\
        }\
    }
    
    
    glBindTexture(tex->target, tex->id);
    
    
    if(tex->use_mipmap)
    {
        /* utilisation du mipmapping */
        
        if(tex->hw_mipmap)
            glTexParameteri(tex->target, GL_GENERATE_MIPMAP_SGIS, GL_TRUE);
        
        if(tex->spec_mipmap)
        {
            /* initialisation des valeurs de mipmap */
            tex->base_level = tex->max_level = tex->data[0].level;
            
            /* donnees de mipmapping envoyees par l'utilisateur */
            for(i=0; i<tex->n_datas; i++)
            {
                SCE_TEXIMAGE(tex->data[i].target, tex->data[i].level,
                             tex->data[i].internal, tex->data[i].w,
                             tex->data[i].h, tex->data[i].d, tex->data[i].size,
                             tex->data[i].format, tex->data[i].type,
                             tex->data[i].texels, tex->data[i].comp)
                
                tex->base_level = (tex->base_level < tex->data[i].level) ?
                                   tex->base_level : tex->data[i].level;
                
                tex->max_level = (tex->max_level > tex->data[i].level) ?
                                  tex->max_level : tex->data[i].level;
            }
        }
        else
        {
            /* l'utilisateur n'a envoye qu'une image.
               ici, il est tout a fait possible qu'il s'agisse que d'une seule
               image, ou bien de six images cubes maps */
            
            if(tex->texcub)
                npass = 6;
            else
                npass = 1;
            
            for(j=0; j<npass; j++)
            {
                /* niveau de base = le niveau de l'image */
                /* NOTE: pour les cubes maps ca risque de chier, car il ne
                         prendra en compte que la derniere texture */
                tex->base_level = tex->data[j].level;
                
                w = tex->data[j].w;
                h = tex->data[j].h;
                d = tex->data[j].d;
                
                size = tex->data[j].size;
                    
                /*     NOTE:--> condition du for douteuse.. <-- */
                for(i=0; w > SCE_MIN_MIPMAP_DIM && h > SCE_MIN_MIPMAP_DIM; i++)
                {
                    SCE_TEXIMAGE(tex->data[j].target, i, tex->data[j].internal,
                                 w, h, d, size, tex->data[j].format,
                                 tex->data[j].type, tex->data[j].texels,
                                 tex->data[j].comp)
                    
                    /* NOTE: calculs douteux */
                    w /= 2; h /= 2; d /= 2;
                    /* re-calcul de la taille */
                    if(tex->tex1d)
                        size =
                        ((w+3)/4) * tex->data[j].pxsize;
                    else if(tex->tex2d || tex->texcub)
                        size =
                        ((w+3)/4) * ((h+3)/4) * tex->data[j].pxsize;
                    else
                        size =
                        ((w+3)/4) * ((h+3)/4) * ((d+3)/4) * tex->data[j].pxsize;
                    
                    /* plus un niveau de mipmapping */
                    tex->max_level++;
                }
            }
        }
    }
    else
    {
        /* pas de mipmapping */
        
        /* pas besoin d'initialiser base_level & max_level,
           ils le sont theoriquement deja */
        
        for(j=0; j<tex->n_datas; j++)
        {
            SCE_TEXIMAGE(tex->data[j].target, 0, tex->data[j].internal,
                         tex->data[j].w, tex->data[j].h, tex->data[j].d,
                         tex->data[j].size, tex->data[j].format,
                         tex->data[j].type, tex->data[j].texels,
                         tex->data[j].comp)
        }
    }
    
    #undef SCE_TEXIMAGE
    
    /* application des niveaux de mipmapping */
    glTexParameteri(tex->target, GL_TEXTURE_BASE_LEVEL, tex->base_level);
    glTexParameteri(tex->target, GL_TEXTURE_MAX_LEVEL, tex->max_level);
    
    glBindTexture(tex->target, 0);
    
    tex->builded = 1;
    
    return (SCE_OK);
}


int SCE_CUpdateTexture(SCE_CTexture *tex)
{
    unsigned int i=0;
    int error = 0;
    
    #ifdef __SCE_DEBUG__
    if(tex == NULL)
    {
        Logger_Log(SCE_INVALID_ARG);
        return (SCE_ERROR);
    }
    #endif
    
    if(tex->builded == 0)
    {
        Logger_Log(SCE_INVALID_OPERATION);
        Logger_LogMsg("you can't update a non-builded texture");
        return (SCE_ERROR);
    }
    
    /* si la texture ne contient aucune donnee */
    if(tex->have_data == 0)
    {
        Logger_Log(SCE_INVALID_OPERATION);
        Logger_LogMsg("you must specify texels before build texture");
        return (SCE_ERROR);
    }
    
    
    /* macro barbare, mais bien pratique :) */
    #define SCE_TEXIMAGE(target, level, w, h, d,\
                         size, format, type, texels, comp)\
    {\
        if(comp)\
        {\
            if(tex->tex1d)\
                glCompressedTexSubImage1D(target, level, 0,\
                                          w, format, size, texels);\
            else if(tex->tex2d)\
                glCompressedTexSubImage2D(target, level, 0, 0,\
                                          w, h, format, size, texels);\
            else if(tex->tex3d)\
                glCompressedTexSubImage3D(target, level, 0, 0, 0,\
                                          w, h, d, format, size, texels);\
        }\
        else\
        {\
            if(tex->tex1d)\
                glTexSubImage1D(target, level, 0,\
                                w, format, type, texels);\
            else if(tex->tex2d)\
                glTexSubImage2D(target, level, 0, 0,\
                                w, h, format, type, texels);\
            else if(tex->tex3d)\
                glTexSubImage3D(target, level, 0, 0, 0,\
                                w, h, d, format, type, texels);\
        }\
    }
    
    /* autre macro barbare :D */
    #define SCE_TEXIMAGE_SWMIPMAP(target, internal, w, h, d,\
                                  format, type, texels)\
    if(tex->tex1d)\
        error = gluBuild1DMipmaps(target, internal, w, format, type, texels);\
    else if(tex->tex2d)\
        error = gluBuild2DMipmaps(target, internal, w, h,\
                                  format, type, texels);\
    else if(tex->tex3d)\
        error = gluBuild3DMipmaps(target, internal, w, h, d,\
                                  format, type, texels);
    
    
    glBindTexture(tex->target, tex->id);
    
    
    /* NOTE: gluBuild*Mipmaps ne supporte pas les formats compresses!!
     *       glu 1.3 les supporte peut-etre... ???
     */
    
    /* mipmapping en hardware ou aucun mipmapping */
    if((tex->use_mipmap && tex->hw_mipmap) || !tex->use_mipmap
       || tex->spec_mipmap )
    {
        for(i=0; i<tex->n_datas; i++)
        {
            SCE_TEXIMAGE(tex->data[i].target, tex->data[i].level,
                         tex->data[i].w, tex->data[i].h, tex->data[i].d,
                         tex->data[i].size, tex->data[i].format,
                         tex->data[i].type, tex->data[i].texels,
                         tex->data[i].comp)
        }
    }
    /* mipmapping software */
    else if(tex->texcub)
    {
        /* cube map */
        for(i=0; i<tex->n_datas; i++)
        {
            SCE_TEXIMAGE_SWMIPMAP(tex->data[i].target, tex->data[i].internal,
                                 tex->data[i].w, tex->data[i].h, tex->data[i].d,
                                 tex->data[i].format, tex->data[i].type,
                                 tex->data[i].texels)
        }
    }
    else
    {
        SCE_TEXIMAGE_SWMIPMAP(tex->data[i].target, tex->data[i].internal,
                              tex->data[i].w, tex->data[i].h, tex->data[i].d,
                              tex->data[i].format, tex->data[i].type,
                              tex->data[i].texels)
    }
    
    if(error != 0)
    {
        /* erreur GLU rencontree, log du message d'erreur */
        Logger_Log(error);
        Logger_LogMsg("function gluBuild*Mipmaps failed: %s",
                      gluErrorString(error));
        
        glBindTexture(tex->target, 0);
        
        return (SCE_ERROR);
    }
    
    glBindTexture(tex->target, 0);
    
    return (SCE_OK);
}


/* revise le 10/07/2007 */
void SCE_CUseTexture(unsigned int unit, SCE_CTexture *tex, unsigned int active)
{
    /* unite de texture non valide */
    if(unit >= SCE_CGetInteger(GL_MAX_TEXTURE_UNITS))
        return;
    
    glActiveTexture(SCE_TEX0 + unit);
    
    if(tex != NULL)
    {
        if(active)
        {
            glEnable(tex->target);
            glBindTexture(tex->target, tex->id);
            binded = tex;
        }
        else
        {
            glBindTexture(tex->target, 0);
            glDisable(tex->target);
            binded = NULL;
        }
    }
    else
    {
        /* aucune texture specifiee, ou desactive tout */
        
        #define SCE_DISABLE_(target)\
        glBindTexture(target, 0);\
        glDisable(target);
        
        SCE_DISABLE_(SCE_TEX_1D)
        SCE_DISABLE_(SCE_TEX_2D)
        SCE_DISABLE_(SCE_TEX_3D)
        SCE_DISABLE_(SCE_TEX_CUBE)
        
        #undef SCE_DISABLE_
        
        binded = NULL;
    }
}
